"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const deps_1 = require("./deps");
const debug = require('debug')('cli-ux');
class CLI {
    constructor(options = {}) {
        this.options = options;
        const globalOptions = global['cli-ux'] || {};
        if (options.mock === undefined)
            options.mock = globalOptions.mock;
        if (options.debug === undefined)
            options.debug = globalOptions.debug;
        this.stdout = new deps_1.deps.StreamOutput(options.mock ? undefined : process.stdout);
        this.stderr = new deps_1.deps.StreamOutput(options.mock ? undefined : process.stderr);
        if (options.mock)
            deps_1.deps.chalk.enabled = false;
    }
    get Prompt() {
        if (!this._prompt) {
            this._prompt = new deps_1.deps.Prompt(this._depOpts);
        }
        return this._prompt;
    }
    get Errors() {
        if (!this._errors) {
            this._errors = new deps_1.deps.Errors(this._depOpts);
        }
        return this._errors;
    }
    get action() {
        if (!this._action) {
            this._action = deps_1.deps.shouldDisplaySpinner(this._depOpts)
                ? new deps_1.deps.SpinnerAction(this._depOpts)
                : new deps_1.deps.SimpleAction(this._depOpts);
        }
        return this._action;
    }
    prompt(name, options = {}) {
        return this.action.pauseAsync(() => {
            return this.Prompt.prompt(name, options);
        }, deps_1.deps.chalk.cyan('?'));
    }
    log(data, ...args) {
        this.action.pause(() => {
            return this.stdout.log(data, ...args);
        });
    }
    warn(err, options = {}) {
        this.action.pause(() => {
            return this.Errors.warn(err, options);
        }, deps_1.deps.chalk.bold.yellow('!'));
    }
    error(err, options = {}) {
        this.action.pause(() => {
            return this.Errors.error(err, options);
        }, deps_1.deps.chalk.bold.red('!'));
    }
    exit(code = 1) {
        this.Errors.exit(code);
    }
    table(data, options) {
        let table = require('./table');
        return table(this, data, options);
    }
    styledJSON(obj) {
        let json = JSON.stringify(obj, null, 2);
        if (deps_1.deps.chalk.enabled) {
            let cardinal = require('cardinal');
            let theme = require('cardinal/themes/jq');
            this.log(cardinal.highlight(json, { json: true, theme: theme }));
        }
        else {
            this.log(json);
        }
    }
    styledHeader(header) {
        this.log(deps_1.deps.chalk.dim('=== ') + deps_1.deps.chalk.bold(header));
    }
    styledObject(obj, keys) {
        const util = require('util');
        let keyLengths = Object.keys(obj).map(key => key.toString().length);
        let maxKeyLength = Math.max.apply(Math, keyLengths) + 2;
        function pp(obj) {
            if (typeof obj === 'string' || typeof obj === 'number') {
                return obj;
            }
            else if (typeof obj === 'object') {
                return Object.keys(obj)
                    .map(k => k + ': ' + util.inspect(obj[k]))
                    .join(', ');
            }
            else {
                return util.inspect(obj);
            }
        }
        let logKeyValue = (key, value) => {
            this.log(`${deps_1.deps.chalk.blue(key)}:` + ' '.repeat(maxKeyLength - key.length - 1) + pp(value));
        };
        for (var key of keys || Object.keys(obj).sort()) {
            let value = obj[key];
            if (Array.isArray(value)) {
                if (value.length > 0) {
                    logKeyValue(key, value[0]);
                    for (var e of value.slice(1)) {
                        this.log(' '.repeat(maxKeyLength) + pp(e));
                    }
                }
            }
            else if (value !== null && value !== undefined) {
                logKeyValue(key, value);
            }
        }
    }
    /**
     * puts in a handler for process.on('uncaughtException') and process.on('unhandledRejection')
     */
    handleUnhandleds() {
        this.Errors.handleUnhandleds();
    }
    /**
     * cleanup any outstanding output like actions that need to be stopped
     */
    done() {
        this.action.stop();
    }
    get _depOpts() {
        return {
            debug: this.options.debug || (this.options.debug === undefined && debug.enabled),
            mock: !!this.options.mock,
            stderr: this.stderr,
            stdout: this.stdout,
        };
    }
}
exports.CLI = CLI;
exports.cli = new CLI();
exports.default = exports.cli;
